--==============================================================================
-- CERN (BE-CO-HT)
-- Xilinx MultiBoot FSM
--==============================================================================
--
-- author: Theodor Stana (t.stana@cern.ch)
--
-- date of creation: 2013-08-19
--
-- version: 1.0
--
-- description: 
--
-- dependencies:
--
-- references:
--  [1]   Xilinx UG380 Spartan-6 FPGA Configuration Guide v2.5
--        http://www.xilinx.com/support/documentation/user_guides/ug380.pdf
-- 
--==============================================================================
-- GNU LESSER GENERAL PUBLIC LICENSE
--==============================================================================
-- This source file is free software; you can redistribute it and/or modify it
-- under the terms of the GNU Lesser General Public License as published by the
-- Free Software Foundation; either version 2.1 of the License, or (at your
-- option) any later version. This source is distributed in the hope that it
-- will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
-- of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
-- See the GNU Lesser General Public License for more details. You should have
-- received a copy of the GNU Lesser General Public License along with this
-- source; if not, download it from http://www.gnu.org/licenses/lgpl-2.1.html
--==============================================================================
-- last changes:
--    2013-08-19   Theodor Stana     t.stana@cern.ch     File created
--==============================================================================
-- TODO: - 
--==============================================================================

library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;


entity multiboot_fsm is
  port
  (
    -- Clock and reset inputs
    clk_i           : in  std_logic;
    rst_n_i         : in  std_logic;

    -- Control register inputs
    rdbootsts_i     : in  std_logic;
    wmb_i           : in  std_logic;
    wgb_i           : in  std_logic;
    iprog_i         : in  std_logic;

    -- Multiboot and golden bitstream start addresses
    gbbar_i         : in  std_logic_vector(31 downto 0);
    mbbar_i         : in  std_logic_vector(31 downto 0);

    -- Outputs to status register
    bootsts_img_o   : out std_logic_vector(15 downto 0);
    bootsts_valid_o : out std_logic;

    -- Ports for the external ICAP component
    icap_dat_i      : in  std_logic_vector(15 downto 0);
    icap_dat_o      : out std_logic_vector(15 downto 0);
    icap_ce_n_o     : out std_logic;
    icap_wr_n_o     : out std_logic
  );
end entity multiboot_fsm;


architecture behav of multiboot_fsm is

  --============================================================================
  -- Type declarations
  --============================================================================
  type t_state is
    (
      IDLE,
      DUMMY_1,
      DUMMY_2,
      SYNC_H,
      SYNC_L,
      SYNC_NOOP_1,
      SYNC_NOOP_2,
      SYNC_NOOP_3,
      SYNC_NOOP_4,
      GEN_1,
      MBA_L,
      GEN_2,
      MBA_H,
      GEN_3,
      GBA_L,
      GEN_4,
      GBA_H,
      IPROG_CMD,
      IPROG,
      BOOTSTS_CMD,
      BOOTSTS_NOOP_1,
      BOOTSTS_NOOP_2,
      BOOTSTS_NOOP_3,
      BOOTSTS_NOOP_4,
      BOOTSTS_SETRD_1,
      BOOTSTS_SETRD_2,
      BOOTSTS_SETRD_3,
      BOOTSTS,
      BOOTSTS_SETWR_1,
      BOOTSTS_SETWR_2,
      BOOTSTS_SETWR_3,
      DESYNC_CMD,
      DESYNC,
      FINAL_NOOP_1,
      FINAL_NOOP_2,
      PREPARE_IDLE
    );

  --============================================================================
  -- Constant declarations
  --============================================================================

  --============================================================================
  -- Component declarations
  --============================================================================

  --============================================================================
  -- Signal declarations
  --============================================================================
  signal state       : t_state;

  signal icap_din    : std_logic_vector(15 downto 0);
  signal icap_dout   : std_logic_vector(15 downto 0);
  signal icap_ce_n   : std_logic;
  signal icap_wr_n   : std_logic;

  signal fsm_cmd     : std_logic_vector(3 downto 0);
  signal fsm_cmd_reg : std_logic_vector(3 downto 0);

--==============================================================================
--  architecture begin
--==============================================================================
begin

  --============================================================================
  -- I/O logic
  --============================================================================
  icap_din    <= icap_dat_i;
  icap_dat_o  <= icap_dout;
  icap_ce_n_o <= icap_ce_n;
  icap_wr_n_o <= icap_wr_n;

  --============================================================================
  -- FSM logic
  --============================================================================
  -- Form state machine command vector from inputs
  fsm_cmd <= iprog_i & wgb_i & wmb_i & rdbootsts_i;

  -- The state machine process
  p_fsm : process(clk_i)
  begin
    if rising_edge(clk_i) then
      if (rst_n_i = '0') then
        state           <= IDLE;
        fsm_cmd_reg     <= (others => '0');
        icap_dout       <= (others => '0');
        icap_ce_n       <= '1';
        icap_wr_n       <= '1';
        bootsts_img_o   <= (others => '0');
        bootsts_valid_o <= '0';
      else

        case state is

          when IDLE =>
            icap_ce_n <= '1';
            icap_wr_n <= '1';
            case fsm_cmd is
              when "1000" | "0001" =>
                state <= DUMMY_1;

              when others =>
                state <= IDLE;

            end case;

            fsm_cmd_reg <= fsm_cmd;

          --====================================================================
          -- Synchronization sequence + four NOOPs
          -- as per Table 6-1, p. 113 [1], steps 1-10
          --====================================================================
          when DUMMY_1 =>
            icap_dout <= x"ffff";
            icap_ce_n <= '0';
            icap_wr_n <= '0';
            state     <= DUMMY_2;

          when DUMMY_2 =>
            icap_dout <= x"ffff";
            icap_ce_n <= '0';
            icap_wr_n <= '0';
            state     <= SYNC_H;

          when SYNC_H =>
            icap_dout <= x"aa99";
            icap_ce_n <= '0';
            icap_wr_n <= '0';
            state     <= SYNC_L;

          when SYNC_L =>
            icap_dout <= x"5566";
            icap_ce_n <= '0';
            icap_wr_n <= '0';
            state     <= SYNC_NOOP_4;

          when SYNC_NOOP_1 =>
            icap_dout <= x"2000";
            icap_ce_n <= '0';
            icap_wr_n <= '0';
            state     <= SYNC_NOOP_2;
            
          when SYNC_NOOP_2 =>
            icap_dout <= x"2000";
            icap_ce_n <= '0';
            icap_wr_n <= '0';
            state     <= SYNC_NOOP_3;

          when SYNC_NOOP_3 =>
            icap_dout <= x"2000";
            icap_ce_n <= '0';
            icap_wr_n <= '0';
            state     <= SYNC_NOOP_4;
            
          when SYNC_NOOP_4 =>
            icap_dout <= x"2000";
            icap_ce_n <= '0';
            icap_wr_n <= '0';
            case fsm_cmd_reg is
              when "1000" =>
                state <= GEN_1;
              when "0001" =>
                state <= BOOTSTS_CMD;
              when others =>
                state <= IDLE;
            end case;

          --====================================================================
          -- IPROG sequence
          -- as per Table 7-1, p. 130 [1], starting from step 4 onward
          --====================================================================
          when GEN_1 =>
            icap_dout <= x"3261";
            icap_ce_n <= '0';
            icap_wr_n <= '0';
            state     <= MBA_L;

          when MBA_L =>
            icap_dout <= mbbar_i(15 downto 0);
            icap_ce_n <= '0';
            icap_wr_n <= '0';
            state     <= GEN_2;

          when GEN_2 =>
            icap_dout <= x"3281";
            icap_ce_n <= '0';
            icap_wr_n <= '0';
            state     <= MBA_H;

          when MBA_H =>
            icap_dout <= mbbar_i(31 downto 16);
            icap_ce_n <= '0';
            icap_wr_n <= '0';
            state     <= GEN_3;

          when GEN_3 =>
            icap_dout <= x"32a1";
            icap_ce_n <= '0';
            icap_wr_n <= '0';
            state     <= GBA_L;

          when GBA_L =>
            icap_dout <= gbbar_i(15 downto 0);
            icap_ce_n <= '0';
            icap_wr_n <= '0';
            state     <= GEN_4;

          when GEN_4 =>
            icap_dout <= x"32c1";
            icap_ce_n <= '0';
            icap_wr_n <= '0';
            state     <= GBA_H;

          when GBA_H =>
            icap_dout <= gbbar_i(31 downto 16);
            icap_ce_n <= '0';
            icap_wr_n <= '0';
            state     <= IPROG_CMD;

          when IPROG_CMD =>
            icap_dout <= x"30a1";
            icap_ce_n <= '0';
            icap_wr_n <= '0';
            state     <= IPROG;

          when IPROG =>
            icap_dout <= x"000e";
            icap_ce_n <= '0';
            icap_wr_n <= '0';
            state     <= FINAL_NOOP_1;

          --====================================================================
          -- BOOTSTS read sequence
          -- as per Table 6-1, p.113 [1], starting from step 6
          --====================================================================
          when BOOTSTS_CMD =>
            icap_ce_n <= '0';
            icap_wr_n <= '0';
            icap_dout <= x"2301";
            state     <= BOOTSTS_NOOP_1;

          when BOOTSTS_NOOP_1 =>
            icap_dout <= x"2000";
            icap_ce_n <= '0';
            icap_wr_n <= '0';
            state     <= BOOTSTS_NOOP_2;

          when BOOTSTS_NOOP_2 =>
            icap_dout <= x"2000";
            icap_ce_n <= '0';
            icap_wr_n <= '0';
            state     <= BOOTSTS_NOOP_3;

          when BOOTSTS_NOOP_3 =>
            icap_dout <= x"2000";
            icap_ce_n <= '0';
            icap_wr_n <= '0';
            state     <= BOOTSTS_NOOP_4;

          when BOOTSTS_NOOP_4 =>
            icap_dout <= x"2000";
            icap_ce_n <= '0';
            icap_wr_n <= '0';
            state     <= BOOTSTS_SETRD_1;

          when BOOTSTS_SETRD_1 =>
            icap_ce_n <= '1';
            icap_wr_n <= '0';
            state     <= BOOTSTS_SETRD_2;

          when BOOTSTS_SETRD_2 =>
            icap_ce_n <= '1';
            icap_wr_n <= '1';
            state     <= BOOTSTS_SETRD_3;

          when BOOTSTS_SETRD_3 =>
            icap_ce_n <= '0';
            icap_wr_n <= '1';
            state     <= BOOTSTS;

          when BOOTSTS =>
            icap_ce_n       <= '0';
            icap_wr_n       <= '1';
            bootsts_img_o   <= icap_din;
            bootsts_valid_o <= '1';
            state           <= BOOTSTS_SETWR_1;

          when BOOTSTS_SETWR_1 =>
            icap_ce_n <= '1';
            icap_wr_n <= '1';
            state     <= BOOTSTS_SETWR_2;

          when BOOTSTS_SETWR_2 =>
            icap_ce_n <= '1';
            icap_wr_n <= '0';
            state     <= BOOTSTS_SETWR_3;

          when BOOTSTS_SETWR_3 =>
            icap_ce_n <= '0';
            icap_wr_n <= '0';
            state     <= DESYNC_CMD;

          when DESYNC_CMD =>
            icap_ce_n <= '0';
            icap_wr_n <= '0';
            icap_dout <= x"30a1";
            state     <= DESYNC;

          when DESYNC =>
            icap_ce_n <= '0';
            icap_wr_n <= '0';
            icap_dout <= x"000d";
            state     <= FINAL_NOOP_1;

          --====================================================================
          -- Two NOOPs at end of all prog sequences
          --====================================================================
          when FINAL_NOOP_1 =>
            icap_ce_n <= '0';
            icap_wr_n <= '0';
            icap_dout <= x"2000";
            state     <= FINAL_NOOP_2; 

          when FINAL_NOOP_2 =>
            icap_ce_n <= '0';
            icap_wr_n <= '0';
            icap_dout <= x"2000";
            state     <= PREPARE_IDLE; 

          --====================================================================
          -- Prepare transition to CE='1', WR='1' in IDLE state
          --====================================================================
          when PREPARE_IDLE =>
            icap_ce_n <= '1';
            icap_wr_n <= '0';
            state     <= IDLE;

          --====================================================================
          -- Safety net
          --====================================================================
          when others =>
            state <= IDLE;

        end case;
      end if;
    end if;
  end process p_fsm;

end architecture behav;
--==============================================================================
--  architecture end
--==============================================================================
