--==============================================================================
-- CERN (BE-CO-HT)
-- Xilinx MultiBoot core top-level file
--==============================================================================
--
-- author: Theodor Stana (t.stana@cern.ch)
--
-- date of creation: 2013-08-19
--
-- version: 1.0
--
-- description: 
--
-- dependencies:
--
-- references:
-- 
--==============================================================================
-- GNU LESSER GENERAL PUBLIC LICENSE
--==============================================================================
-- This source file is free software; you can redistribute it and/or modify it
-- under the terms of the GNU Lesser General Public License as published by the
-- Free Software Foundation; either version 2.1 of the License, or (at your
-- option) any later version. This source is distributed in the hope that it
-- will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
-- of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
-- See the GNU Lesser General Public License for more details. You should have
-- received a copy of the GNU Lesser General Public License along with this
-- source; if not, download it from http://www.gnu.org/licenses/lgpl-2.1.html
--==============================================================================
-- last changes:
--    2013-08-19   Theodor Stana     t.stana@cern.ch     File created
--==============================================================================
-- TODO: - 
--==============================================================================

library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

library unisim;
use unisim.vcomponents.all;

use work.wishbone_pkg.all;

entity xil_multiboot is
  port
  (
    -- Clock and reset input ports
    clk_i   : in  std_logic;
    rst_n_i : in  std_logic;

    -- Wishbone ports
    wbs_i   : in  t_wishbone_slave_in;
    wbs_o   : out t_wishbone_slave_out
  );
end entity xil_multiboot;


architecture behav of xil_multiboot is

  --============================================================================
  -- Type declarations
  --============================================================================

  --============================================================================
  -- Constant declarations
  --============================================================================

  --============================================================================
  -- Component declarations
  --============================================================================
  -- Register component
  component multiboot_regs is
    port (
      -- Clock and reset inputs
      rst_n_i            : in     std_logic;
      clk_sys_i          : in     std_logic;

      -- Wishbone ports
      wb_adr_i           : in     std_logic_vector(1 downto 0);
      wb_dat_i           : in     std_logic_vector(31 downto 0);
      wb_dat_o           : out    std_logic_vector(31 downto 0);
      wb_cyc_i           : in     std_logic;
      wb_sel_i           : in     std_logic_vector(3 downto 0);
      wb_stb_i           : in     std_logic;
      wb_we_i            : in     std_logic;
      wb_ack_o           : out    std_logic;
      wb_stall_o         : out    std_logic;

      -- Fields of control register
      multiboot_cr_rdbootsts_o     : out    std_logic;
      multiboot_cr_wmb_o           : out    std_logic;
      multiboot_cr_wgb_o           : out    std_logic;
      multiboot_cr_iprog_o         : out    std_logic;

      -- Fields of status register
      multiboot_sr_bootsts_img_i   : in std_logic_vector(15 downto 0);
      multiboot_sr_valid_i         : in std_logic;

      -- Fields of bitstream address registers
      multiboot_gbbar_o            : out std_logic_vector(31 downto 0);
      multiboot_mbbar_o            : out std_logic_vector(31 downto 0)
    );
  end component multiboot_regs;

  -- FSM component
  component multiboot_fsm is
    port
    (
      -- Clock and reset inputs
      clk_i   : in  std_logic;
      rst_n_i : in  std_logic;

      -- Control register inputs
      rdbootsts_i : in  std_logic;
      wmb_i       : in  std_logic;
      wgb_i       : in  std_logic;
      iprog_i     : in  std_logic;

      -- Multiboot and golden bitstream start addresses
      gbbar_i         : in  std_logic_vector(31 downto 0);
      mbbar_i         : in  std_logic_vector(31 downto 0);

      -- Outputs to status register
      bootsts_img_o   : out std_logic_vector(15 downto 0);
      bootsts_valid_o : out std_logic;

      -- Data input and outputs for ICAP component
      icap_dat_i      : in  std_logic_vector(15 downto 0);
      icap_dat_o      : out std_logic_vector(15 downto 0);

      -- Active low chip- and write-enable outputs for ICAP component
      icap_ce_n_o     : out std_logic;
      icap_wr_n_o     : out std_logic
    );
  end component multiboot_fsm;

  --============================================================================
  -- Signal declarations
  --============================================================================
  -- Control and status register signals
  signal rdbootsts     : std_logic;
  signal wmb, wgb      : std_logic;
  signal iprog         : std_logic;
  signal bootsts_img   : std_logic_vector(15 downto 0);
  signal sr_valid      : std_logic;
  signal gbbar, mbbar  : std_logic_vector(31 downto 0);

  -- ICAP signals
  signal icap_ce_n     : std_logic;
  signal icap_wr_n     : std_logic;
  signal icap_busy     : std_logic;
  signal icap_din      : std_logic_vector(15 downto 0);
  signal icap_dout     : std_logic_vector(15 downto 0);

  -- FSM signals
  signal fsm_icap_din  : std_logic_vector(15 downto 0);
  signal fsm_icap_dout : std_logic_vector(15 downto 0);

--==============================================================================
--  architecture begin
--==============================================================================
begin

  --============================================================================
  -- Register component instantiation
  --============================================================================
  cmp_regs : multiboot_regs
    port map
    (
      rst_n_i                    => rst_n_i,
      clk_sys_i                  => clk_i,

      wb_adr_i                   => wbs_i.adr(3 downto 2),
      wb_dat_i                   => wbs_i.dat,
      wb_dat_o                   => wbs_o.dat,
      wb_cyc_i                   => wbs_i.cyc,
      wb_sel_i                   => wbs_i.sel,
      wb_stb_i                   => wbs_i.stb,
      wb_we_i                    => wbs_i.we,
      wb_ack_o                   => wbs_o.ack,
      wb_stall_o                 => wbs_o.stall,

      multiboot_cr_rdbootsts_o   => rdbootsts,
      multiboot_cr_wmb_o         => wmb,
      multiboot_cr_wgb_o         => wgb,
      multiboot_cr_iprog_o       => iprog,

      multiboot_sr_bootsts_img_i => bootsts_img,
      multiboot_sr_valid_i       => sr_valid,

      multiboot_gbbar_o          => gbbar,
      multiboot_mbbar_o          => mbbar
    );

  --============================================================================
  -- FSM component instantiation
  --============================================================================
  cmp_fsm : multiboot_fsm
    port map
    (
      clk_i           => clk_i,
      rst_n_i         => rst_n_i,

      rdbootsts_i     => rdbootsts,
      wmb_i           => wmb,
      wgb_i           => wgb,
      iprog_i         => iprog,

      gbbar_i         => gbbar,
      mbbar_i         => mbbar,

      bootsts_img_o   => bootsts_img,
      bootsts_valid_o => sr_valid,

      icap_dat_i      => fsm_icap_din,
      icap_dat_o      => fsm_icap_dout,

      icap_ce_n_o     => icap_ce_n,
      icap_wr_n_o     => icap_wr_n
    );

  --============================================================================
  -- Xilinx ICAP logic
  --============================================================================
  -- First, bit-flip the data to/from the FSM
  icap_din( 0) <= fsm_icap_dout( 7);
  icap_din( 1) <= fsm_icap_dout( 6);
  icap_din( 2) <= fsm_icap_dout( 5);
  icap_din( 3) <= fsm_icap_dout( 4);
  icap_din( 4) <= fsm_icap_dout( 3);
  icap_din( 5) <= fsm_icap_dout( 2);
  icap_din( 6) <= fsm_icap_dout( 1);
  icap_din( 7) <= fsm_icap_dout( 0);
  icap_din( 8) <= fsm_icap_dout(15);
  icap_din( 9) <= fsm_icap_dout(14);
  icap_din(10) <= fsm_icap_dout(13);
  icap_din(11) <= fsm_icap_dout(12);
  icap_din(12) <= fsm_icap_dout(11);
  icap_din(13) <= fsm_icap_dout(10);
  icap_din(14) <= fsm_icap_dout( 9);
  icap_din(15) <= fsm_icap_dout( 8);

  fsm_icap_din( 0) <= icap_dout( 7);
  fsm_icap_din( 1) <= icap_dout( 6);
  fsm_icap_din( 2) <= icap_dout( 5);
  fsm_icap_din( 3) <= icap_dout( 4);
  fsm_icap_din( 4) <= icap_dout( 3);
  fsm_icap_din( 5) <= icap_dout( 2);
  fsm_icap_din( 6) <= icap_dout( 1);
  fsm_icap_din( 7) <= icap_dout( 0);
  fsm_icap_din( 8) <= icap_dout(15);
  fsm_icap_din( 9) <= icap_dout(14);
  fsm_icap_din(10) <= icap_dout(13);
  fsm_icap_din(11) <= icap_dout(12);
  fsm_icap_din(12) <= icap_dout(11);
  fsm_icap_din(13) <= icap_dout(10);
  fsm_icap_din(14) <= icap_dout( 9);
  fsm_icap_din(15) <= icap_dout( 8);

  -- and instantiate the ICAP component
  cmp_icap : ICAP_SPARTAN6
    port map
    (
      CLK   => clk_i,
      CE    => icap_ce_n,
      WRITE => icap_wr_n,
      I     => icap_din,
      O     => icap_dout,
      BUSY  => icap_busy
    );

end architecture behav;
--==============================================================================
--  architecture end
--==============================================================================
