--==============================================================================
-- CERN (BE-CO-HT)
-- Testbench for Xilinx MultiBoot design
--==============================================================================
--
-- author: Theodor Stana (t.stana@cern.ch)
--
-- date of creation: 2013-08-19
--
-- version: 1.0
--
-- description: 
--
-- dependencies:
--
-- references:
-- 
--==============================================================================
-- GNU LESSER GENERAL PUBLIC LICENSE
--==============================================================================
-- This source file is free software; you can redistribute it and/or modify it
-- under the terms of the GNU Lesser General Public License as published by the
-- Free Software Foundation; either version 2.1 of the License, or (at your
-- option) any later version. This source is distributed in the hope that it
-- will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
-- of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
-- See the GNU Lesser General Public License for more details. You should have
-- received a copy of the GNU Lesser General Public License along with this
-- source; if not, download it from http://www.gnu.org/licenses/lgpl-2.1.html
--==============================================================================
-- last changes:
--    2013-08-19   Theodor Stana     t.stana@cern.ch     File created
--==============================================================================
-- TODO: - 
--==============================================================================

library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

use std.textio.all;
use work.wishbone_pkg.all;

entity testbench is
end entity testbench;


architecture behav of testbench is

  --============================================================================
  -- Type declarations
  --============================================================================

  --============================================================================
  -- Constant declarations
  --============================================================================
  constant c_clk_per   : time := 8 ns;
  constant c_rst_width : time := 130 ns;

  --============================================================================
  -- Component declarations
  --============================================================================
  component xil_multiboot is
    port
    (
      -- Clock and reset input ports
      clk_i   : in  std_logic;
      rst_n_i : in  std_logic;

      -- Wishbone ports
      wbs_i   : in  t_wishbone_slave_in;
      wbs_o   : out t_wishbone_slave_out
    );
  end component xil_multiboot;

  component multiboot_regs is
    port (
      -- Clock and reset inputs
      rst_n_i            : in     std_logic;
      clk_sys_i          : in     std_logic;
  
      -- Wishbone ports
      wb_adr_i           : in     std_logic_vector(1 downto 0);
      wb_dat_i           : in     std_logic_vector(31 downto 0);
      wb_dat_o           : out    std_logic_vector(31 downto 0);
      wb_cyc_i           : in     std_logic;
      wb_sel_i           : in     std_logic_vector(3 downto 0);
      wb_stb_i           : in     std_logic;
      wb_we_i            : in     std_logic;
      wb_ack_o           : out    std_logic;
      wb_stall_o         : out    std_logic;
  
      -- Fields of control register
      multiboot_cr_rdbootsts_o     : out    std_logic;
      multiboot_cr_wmb_o           : out    std_logic;
      multiboot_cr_wgb_o           : out    std_logic;
      multiboot_cr_iprog_o         : out    std_logic;
  
      -- Fields of status register
      multiboot_sr_bootsts_img_i   : in std_logic_vector(15 downto 0);
      multiboot_sr_valid_i         : in std_logic;
  
      -- Fields of bitstream address registers
      multiboot_gbbar_o              : out std_logic_vector(31 downto 0);
      multiboot_mbbar_o              : out std_logic_vector(31 downto 0)
    );
  end component multiboot_regs;

  --============================================================================
  -- Signal declarations
  --============================================================================
  signal rst_n, clk   : std_logic := '0';

  signal wb_adr       : std_logic_vector(31 downto 0);
  signal wb_dat_in    : std_logic_vector(31 downto 0);
  signal wb_dat_out   : std_logic_vector(31 downto 0);
  signal wb_cyc       : std_logic;
  signal wb_stb       : std_logic;
  signal wb_we        : std_logic;
  signal wb_ack       : std_logic;
  signal wb_stall     : std_logic;
  signal wb_sel       : std_logic_vector(3 downto 0);

  signal wbs_in       : t_wishbone_slave_in;
  signal wbs_out      : t_wishbone_slave_out;

  signal transfer     : std_logic;
  signal write        : std_logic;
  signal adr          : std_logic_vector(31 downto 0);
  signal dat          : std_logic_vector(31 downto 0);

  signal str : string(1 to 8);

--==============================================================================
--  architecture begin
--==============================================================================
begin

  --============================================================================
  -- Instantiate UUT
  --============================================================================
  UUT: xil_multiboot
    port map
    (
      rst_n_i => rst_n,
      clk_i   => clk,
      wbs_i   => wbs_in,
      wbs_o   => wbs_out
    );

  -- bind Wishbone ports to signals
  wbs_in.dat <= wb_dat_out;
  wbs_in.adr <= wb_adr;
  wbs_in.cyc <= wb_cyc;
  wbs_in.stb <= wb_stb;
  wbs_in.we  <= wb_we;
  wb_ack     <= wbs_out.ack;
  wb_dat_in  <= wbs_out.dat;
  wb_stall   <= wbs_out.stall;

  --============================================================================
  -- Clock and reset processes
  --============================================================================
  p_clk : process is
  begin
    clk <= not clk;
    wait for c_clk_per/2;
  end process p_clk;

  p_rst : process is
  begin
    rst_n <= '0';
    wait for c_rst_width;
    rst_n <= '1';
    wait;
  end process;

  --============================================================================
  -- Stimuli process
  --============================================================================
  p_stim : process is
  begin

    str <= "init    ";
    adr <= x"00000000";
    dat <= x"00000000";
    write <= '0';
    transfer <= '0';

    wait for 1500ns;

    -- Send IPROG
    wait for 200 ns;
    str <= "wr-iprog";
    adr <= x"00000000";
    dat <= x"00000008";
    write <= '1';
    transfer <= '1';
    wait for c_clk_per;
    transfer <= '0';

    -- Send RDBOOTSTS
    wait for 200 ns;
    str <= "wr-bsts ";
    adr <= x"00000000";
    dat <= x"00000001";
    write <= '1';
    transfer <= '1';
    wait for c_clk_per;
    transfer <= '0';

    -- Read from SR
    wait for 500 ns;
    str <= "rd-sr   ";
    adr <= x"00000004";
    write <= '0';
    transfer <= '1';
    wait for c_clk_per;
    transfer <= '0';

    ---- Read from CR
    --wait for 30 ns;
    --str <= "rd-cr   ";
    --adr <= x"00000000";
    --write <= '0';
    --transfer <= '1';
    --wait for c_clk_per;
    --transfer <= '0';

    ---- Read from SR
    --wait for 30 ns;
    --str <= "rd-sr   ";
    --valid <= '1';
    --bootsts_img <= x"f3f3";
    --adr <= x"00000004";
    --write <= '0';
    --transfer <= '1';
    --wait for c_clk_per;
    --transfer <= '0';

    ---- write to GBBAR
    --wait for 30 ns;
    --str <= "wr-gbbar";
    --adr <= x"00000008";
    --dat <= x"0b012345";
    --write <= '1';
    --transfer <= '1';
    --wait for c_clk_per;
    --transfer <= '0';

    ---- read from GBBAR
    --wait for 30 ns;
    --str <= "rd-gbbar";
    --adr <= x"00000008";
    --write <= '0';
    --transfer <= '1';
    --wait for c_clk_per;
    --transfer <= '0';

    ---- write to MBBAR
    --wait for 30 ns;
    --str <= "wr-mbbar";
    --adr <= x"0000000C";
    --dat <= x"0b024321";
    --write <= '1';
    --transfer <= '1';
    --wait for c_clk_per;
    --transfer <= '0';

    ---- read from MBBAR
    --wait for 30 ns;
    --str <= "rd-mbbar";
    --adr <= x"0000000C";
    --write <= '0';
    --transfer <= '1';
    --wait for c_clk_per;
    --transfer <= '0';

    ---- Read from SR
    --wait for 30 ns;
    --str <= "rd-sr   ";
    --valid <= '1';
    --bootsts_img <= x"3f3f";
    --adr <= x"00000004";
    --write <= '0';
    --transfer <= '1';
    --wait for c_clk_per;
    --transfer <= '0';

    -- wait indefinitely
    wait;

  end process p_stim;

  --============================================================================
  -- Process running the Wishbone master
  --============================================================================
  p_wbm : process (clk) is
  begin
    if rising_edge(clk) then
      if (rst_n = '0') then
        wb_adr     <= (others => '0');
        wb_dat_out <= (others => '0');
        wb_sel     <= (others => '0');
        wb_stb     <= '0';
        wb_cyc     <= '0';
        wb_we      <= '0';
      else
        if (transfer = '1') then
          wb_adr     <= adr;
          wb_dat_out <= dat;
          wb_we      <= write;
          wb_stb     <= '1';
          wb_cyc     <= '1';
        elsif (wb_ack = '1') then
          wb_stb <= '0';
          wb_cyc <= '0';
        end if;
      end if;
    end if;
  end process p_wbm;

end architecture behav;
--==============================================================================
--  architecture end
--==============================================================================
